################################################################################
# SpaceRocks, version 1.0.11
#
# Author: Kevin Napier kjnapier@umich.edu
################################################################################
import random
import copy
import os

from astropy import units as u
from astropy.coordinates import Angle, Distance
from astropy.time import Time

from numpy import sin, cos, arctan2, sqrt, array, pi, zeros, sinh, arctanh, arccosh, cosh, tanh, zeros_like
import numpy as np
import pandas as pd

import rebound
from rebound import hash as h

#import reboundx
#from reboundx import constants
from scipy.optimize import newton

from .constants import *
from .orbitfuncs import OrbitFuncs
from .convenience import Convenience
from .units import Units
from .vector import Vector
from .ephemerides import Ephemerides

from skyfield.api import Topos, Loader
# Load in planets for ephemeride calculation.
load = Loader('./Skyfield-Data', expire=False, verbose=False)
ts = load.timescale()
planets = load('de440.bsp')
earth = planets['earth']
sun = planets['sun']

import pkg_resources

DATA_PATH = pkg_resources.resource_filename('spacerocks', 'data/observatories.csv')


#bservatories = pd.read_csv(os.path.join(os.path.dirname(__file__),
#                            'data',
#                            'observatories.csv'))

observatories = pd.read_csv(DATA_PATH)

from skyfield.api import wgs84
from skyfield.data import iers

ts = load.timescale()

import ctypes
from numpy.ctypeslib import ndpointer

from . import clibspacerocks

clibspacerocks.kep_to_xyz_temp.argtypes = [ctypes.c_int,
                                   ndpointer(ctypes.c_double, flags='C_CONTIGUOUS'),
                                   ndpointer(ctypes.c_double, flags='C_CONTIGUOUS'),
                                   ndpointer(ctypes.c_double, flags='C_CONTIGUOUS'),
                                   ndpointer(ctypes.c_double, flags='C_CONTIGUOUS'),
                                   ndpointer(ctypes.c_double, flags='C_CONTIGUOUS'),
                                   ndpointer(ctypes.c_double, flags='C_CONTIGUOUS')]

clibspacerocks.kep_to_xyz_temp.restype = ctypes.POINTER(ctypes.c_double)

def kep_to_xyz_temp_cpp(N, a, e, inc, arg, node, M):

    rock = clibspacerocks.kep_to_xyz_temp(N, a, e, inc, arg, node, M)
    arr = np.ctypeslib.as_array(rock, (6 * N,))

    x, y, z, vx, vy, vz = arr.reshape(N, 6).T
    return x * u.au, y * u.au, z * u.au, vx * u.au/u.d, vy * u.au/u.d, vz * u.au/u.d

class SpaceRock(OrbitFuncs, Convenience):

    def __init__(self, frame='barycentric', units=Units(), *args, **kwargs):

        coords = self.detect_coords(kwargs)
        frame = frame.lower()

        # input -> arrays
        for idx, key in enumerate([*kwargs]):
            kwargs[key] = np.atleast_1d(kwargs.get(key))


        self.frame = frame
        if self.frame == 'barycentric':
            self.mu = mu_bary
        elif self.frame == 'heliocentric':
            self.mu = mu_helio


        if coords == 'kep':

            if kwargs.get('a') is not None:
                self.a = Distance(kwargs.get('a'), units.distance, allow_negative=True)

            if kwargs.get('b') is not None:
                self.b = Distance(kwargs.get('b'), units.distance, allow_negative=True)

            if kwargs.get('v_inf') is not None:
                self.v_inf = (kwargs.get('v_inf') * units.speed).to(u.au / u.day)

            if kwargs.get('e') is not None:
                self.e = kwargs.get('e') #* u.dimensionless_unscaled

            self.inc = Angle(kwargs.get('inc'), units.angle)

            if kwargs.get('q') is not None:
                self.q = Distance(kwargs.get('q'), units.distance)

            if kwargs.get('Q') is not None:
                self.Q = Distance(kwargs.get('Q'), units.distance)

            if kwargs.get('node') is not None:
                self.node = Angle(kwargs.get('node'), units.angle)

            if kwargs.get('arg') is not None:
                self.arg = Angle(kwargs.get('arg'), units.angle)

            if kwargs.get('varpi') is not None:
                self.varpi = Angle(kwargs.get('varpi'), units.angle)

            if kwargs.get('t_peri') is not None:
                if units.timeformat is None:
                    self.t_peri = self.detect_timescale(kwargs.get('t_peri'), units.timescale)
                else:
                    self.t_peri = Time(kwargs.get('t_peri'), format=units.timeformat, scale=units.timescale)

            if kwargs.get('M') is not None:
                self.M = Angle(kwargs.get('M'), units.angle)
                #self.M[(self.e > 1) & (self.M.rad > np.pi)] -= 2 * np.pi * u.rad

            if kwargs.get('E') is not None:
                self.E = Angle(kwargs.get('E'), units.angle)

            if kwargs.get('true_anomaly') is not None:
                self.true_anomaly = Angle(kwargs.get('true_anomaly'), units.angle)

            if kwargs.get('true_longitude') is not None:
                self.true_longitude = Angle(kwargs.get('true_longitude'), units.angle)

            if kwargs.get('mean_longitude') is not None:
                self.mean_longitude = Angle(kwargs.get('mean_longitude'), units.angle)


            if kwargs.get('epoch') is not None:

                if units.timeformat is None:
                    self.epoch = self.detect_timescale(kwargs.get('epoch'), units.timescale)
                else:
                    self.epoch = Time(kwargs.get('epoch'), format=units.timeformat, scale=units.timescale)
            else:

                self.epoch = Time(np.zeros(len(self.inc)), format='jd', scale='utc')

            if kwargs.get('name') is not None:
                self.name = np.atleast_1d(kwargs.get('name'))
            else:
                # produces random, non-repeting integers between 0 and 1e10 - 1
                self.name = array(['{:010}'.format(value) for value in random.sample(range(int(1e10)), len(self.inc))])


        elif coords == 'xyz':

            x = Distance(kwargs.get('x'), units.distance, allow_negative=True)
            y = Distance(kwargs.get('y'), units.distance, allow_negative=True)
            z = Distance(kwargs.get('z'), units.distance, allow_negative=True)
            vx = (kwargs.get('vx') * units.speed).to(u.au / u.day)
            vy = (kwargs.get('vy') * units.speed).to(u.au / u.day)
            vz = (kwargs.get('vz') * units.speed).to(u.au / u.day)

            self.position = Vector(x, y, z)
            self.velocity = Vector(vx, vy, vz)

            if kwargs.get('epoch') is not None:

                if units.timeformat is None:
                    self.epoch = self.detect_timescale(kwargs.get('epoch'), units.timescale)
                else:
                    self.epoch = Time(kwargs.get('epoch'), format=units.timeformat, scale=units.timescale)
            else:
                self.epoch = Time(np.zeros(len(self.x)), format='jd', scale='utc')


            if kwargs.get('name') is not None:
                self.name = np.atleast_1d(kwargs.get('name'))
            else:
                # produces random, non-repeting integers between 0 and 1e10 - 1
                self.name = array(['{:010}'.format(value) for value in random.sample(range(int(1e10)), len(self.x))])


        if kwargs.get('H0') is not None:
            self.H0 = kwargs.get('H0')


        if kwargs.get('G') is not None:
            self.G = kwargs.get('G')
        else:
            self.G = np.repeat(0.15, len(self))


        if kwargs.get('mag') is not None:
            self.mag = kwargs.get('mag')


        if (kwargs.get('rotation_period') is not None) and (kwargs.get('delta_H') is not None) and (kwargs.get('phi0') is not None):
            self.rotation_period = kwargs.get('rotation_period')
            self.delta_H = kwargs.get('delta_H')
            self.phi0 = Angle(kwargs.get('phi0'), units.angle)
            self.t0 = Time(self.epoch.jd, format='jd', scale=units.timescale)


    def analytic_propagate(self, epoch, propagate_frame = 'heliocentric'):
        '''
        propagate all bodies to the desired date using Kieprian orbit.
        '''
        in_frame = self.frame
        if propagate_frame != in_frame:
            if propagate_frame == 'heliocentric':
                self.to_helio()
            else:
                self.to_bary()

        M = (self.n.value * (epoch - self.epoch.jd) + self.M.rad*180/np.pi)%360

        rocks = SpaceRock(a=self.a,
                          e=self.e,
                          inc=self.inc,
                          node=self.node,
                          arg=self.arg,
                          M=M,
                          name=self.name,
                          epoch=epoch,
                          frame=propagate_frame)

        # be polite and return orbital parameters in the input frame.
        if in_frame != self.frame:
            if in_frame == 'heliocentric':
                self.to_helio()
            else:
                self.to_bary()

        if hasattr(self, 'G'):
            rocks.G = self.G

        if hasattr(self, 'mag'):
            rocks.mag = self.mag

        if hasattr(self, 'delta_H'):
            rocks.delta_H = self.delta_H
            rocks.rotation_period = self.rotation_period
            rocks.phi0 = self.phi0
            rocks.t0 = Time(self.t0.jd, format='jd')

            rocks.H = self.H + rocks.delta_H * np.sin(2 * np.pi * (rocks.epoch.jd - rocks.t0.jd) / rocks.rotation_period  - rocks.phi0)

        elif hasattr(self, 'H'):
            rocks.H = self.H

        return rocks

    def propagate(self, epochs, model, units=Units()):
        '''
        Integrate all bodies to the desired date. The logic could be cleaner
        but it works.
        '''

        epochs = self.detect_timescale(np.atleast_1d(epochs), units.timescale)

        Nx = len(epochs)

        frame = self.frame

        # We need to integrate in barycentric coordinates
        self.to_bary()

        # Integrate all particles to the same obsdate
        pickup_times = self.epoch.tdb.jd
        sim, planet_names = self.set_simulation(np.min(pickup_times), units=units, model=model)
        sim.t = np.min(pickup_times) #np.min(df.epoch)

        # need to ensure these are computed
        self.x
        self.y
        self.z
        self.vx
        self.vy
        self.vz

        for time in np.sort(np.unique(pickup_times)):
            ps = self[self.epoch.tdb.jd == time]
            for x, y, z, vx, vy, vz, name in zip(ps.x.value, ps.y.value, ps.z.value, ps.vx.value, ps.vy.value, ps.vz.value, ps.name):
                sim.add(x=x, y=y, z=z, vx=vx, vy=vy, vz=vz, hash=name)
                sim.integrate(time, exact_finish_time=1)

        sim.move_to_com()


        x_values = np.zeros((Nx, sim.N))
        y_values = np.zeros((Nx, sim.N))
        z_values = np.zeros((Nx, sim.N))
        vx_values = np.zeros((Nx, sim.N))
        vy_values = np.zeros((Nx, sim.N))
        vz_values = np.zeros((Nx, sim.N))
        name_values = np.zeros((Nx, sim.N), dtype='object')
        obsdate_values = np.zeros((Nx, sim.N))

        p_names = planet_names + [name for name in self.name]

        for ii, time in enumerate(np.sort(epochs.tdb.jd)):
            sim.integrate(time, exact_finish_time=1)
            a = np.zeros((sim.N, 3), dtype='float64')
            b = np.zeros((sim.N, 3), dtype='float64')
            sim.serialize_particle_data(xyz=a, vxvyvz=b)

            x, y, z = a.T
            vx, vy, vz = b.T

            x_values[ii] = x
            y_values[ii] = y
            z_values[ii] = z
            vx_values[ii] = vx
            vy_values[ii] = vy
            vz_values[ii] = vz
            name_values[ii] = p_names
            obsdate_values[ii] = np.repeat(sim.t, sim.N)


        x = x_values[:, sim.N_active:].flatten()
        y = y_values[:, sim.N_active:].flatten()
        z = z_values[:, sim.N_active:].flatten()
        vx = vx_values[:, sim.N_active:].flatten()
        vy = vy_values[:, sim.N_active:].flatten()
        vz = vz_values[:, sim.N_active:].flatten()
        name = name_values[:, sim.N_active:].flatten()
        epoch = obsdate_values[:, sim.N_active:].flatten()

        px = x_values[:, :sim.N_active].flatten()
        py = y_values[:, :sim.N_active].flatten()
        pz = z_values[:, :sim.N_active].flatten()
        pvx = vx_values[:, :sim.N_active].flatten()
        pvy = vy_values[:, :sim.N_active].flatten()
        pvz = vz_values[:, :sim.N_active].flatten()
        pname = name_values[:, :sim.N_active].flatten()
        pepoch = obsdate_values[:, :sim.N_active].flatten()


        units = Units()
        units.timescale = 'tdb'
        rocks = self.__class__(x=x, y=y, z=z, vx=vx, vy=vy, vz=vz, name=name, epoch=epoch, frame='barycentric', units=units)
        planets = self.__class__(x=px, y=py, z=pz, vx=pvx, vy=pvy, vz=pvz, name=pname, epoch=pepoch, frame='barycentric', units=units)

        # be polite and return orbital parameters in the input frame.
        if frame == 'heliocentric':
            rocks.to_helio()
            self.to_helio()


        if hasattr(self, 'G'):
            rocks.G = np.tile(self.G, Nx)


        if hasattr(self, 'delta_H'):
            rocks.delta_H = np.tile(self.delta_H, Nx)
            rocks.rotation_period = np.tile(self.rotation_period, Nx)
            rocks.phi0 = np.tile(self.phi0, Nx)
            rocks.t0 = Time(np.tile(self.t0.jd, Nx), format='jd')
            rocks.H0 = np.tile(self.H0, Nx)


        elif hasattr(self, 'H0'):
            rocks.H0 = np.tile(self.H0, Nx)

        return rocks, planets, sim

    def calc_H(self, obscode):
        obs = self.observe(obscode=obscode)

        t = ts.tdb(jd=self.epoch.tdb.jd)
        e = earth.at(t)
        s = sun.at(t)

        sx, sy, sz = s.ecliptic_xyz().au
        ex, ey, ez = e.ecliptic_xyz().au
        earth_dist = ((ex-sx)**2 + (ey-sy)**2 + (ez-sz)**2)**0.5

        q = (obs.r_helio.au**2 + obs.delta.au**2 - earth_dist**2)/(2 * obs.r_helio.au * obs.delta.au)

        beta = np.arccos(q)
        beta[np.where(q <= -1)[0]] = np.pi * u.rad
        beta[np.where(q >= 1)[0]] = 0 * u.rad

        Psi_1 = np.exp(-3.332 * np.tan(beta/2)**0.631)
        Psi_2 = np.exp(-1.862 * np.tan(beta/2)**1.218)


        H = self.mag - 5 * np.log10(obs.r_helio.au * obs.delta.au)# / earth_dist**2)

        not_zero = np.where((Psi_1 != 0) | (Psi_2 != 0))[0]
        H[not_zero] += 2.5 * np.log10((1 - self.G[not_zero]) * Psi_1[not_zero] + self.G[not_zero] * Psi_2[not_zero])

        return H


    def observe(self, obscode):

        if self.frame == 'barycentric':
            self.to_helio()
            r_helio = self.r
            self.to_bary()
        else:
            r_helio = self.r

        x, y, z, vx, vy, vz = self.xyz_to_tel(obscode)

        if not hasattr(self, 'H0'):
            return Ephemerides(x=x, y=y, z=z, vx=vx, vy=vy, vz=vz, epoch=self.epoch, name=self.name, r_helio=r_helio)
        elif (not hasattr(self, 'delta_H')) and hasattr(self, 'H0'):
            return Ephemerides(x=x, y=y, z=z, vx=vx, vy=vy, vz=vz, epoch=self.epoch, name=self.name, r_helio=r_helio, H0=self.H0, G=self.G)
        else:
            return Ephemerides(x=x, y=y, z=z, vx=vx, vy=vy, vz=vz, epoch=self.epoch, name=self.name, r_helio=r_helio, H0=self.H0, G=self.G, delta_H=self.delta_H, t0=self.t0, rotation_period=self.rotation_period, phi0=self.phi0)

    def xyz_to_tel(self, obscode=500):
        '''
        Transform from barycentric ecliptic Cartesian coordinates to
        telescope-centric coordinates.

        Routine corrects iteratively for light travel time.
        '''

        if obscode != 500:
            obscode = str(obscode).zfill(3)
            obs = observatories[observatories.obscode == obscode]
            obslat = obs.lat.values
            obslon = obs.lon.values
            obselev = obs.elevation.values


        in_frame = self.frame
        #if in_frame == 'heliocentric':
        self.to_bary()


        alltimes = self.epoch.tdb.jd
        unique_times = np.unique(alltimes)
        t = ts.tdb(jd=unique_times)
        earth = planets['earth']

        if obscode == 'ssb':
            x_observer = np.zeros(len(alltimes)) * u.au
            y_observer = np.zeros(len(alltimes)) * u.au
            z_observer = np.zeros(len(alltimes)) * u.au

            vx_observer = np.zeros(len(alltimes)) * u.au / u.day
            vy_observer = np.zeros(len(alltimes)) * u.au / u.day
            vz_observer = np.zeros(len(alltimes)) * u.au / u.day

        elif obscode == 'sun':
            ss = sun.at(t)

            xx, yy, zz = ss.position.au * u.au # earth ICRS position
            vxx, vyy, vzz = ss.velocity.au_per_d * u.au / u.day # earth ICRS position


            sxs = {t:x.value for t, x in zip(unique_times, xx)}
            sys = {t:y.value for t, y in zip(unique_times, yy)}
            szs = {t:z.value for t, z in zip(unique_times, zz)}
            svxs = {t:vx.value for t, vx in zip(unique_times, vxx)}
            svys = {t:vy.value for t, vy in zip(unique_times, vyy)}
            svzs = {t:vz.value for t, vz in zip(unique_times, vzz)}

            x_observer = np.array([sxs[t] for t in alltimes]) * u.au
            y_observer = np.array([sys[t] for t in alltimes]) * u.au
            z_observer = np.array([szs[t] for t in alltimes]) * u.au

            vx_observer = np.array([svxs[t] for t in alltimes]) * u.au / u.day
            vy_observer = np.array([svys[t] for t in alltimes]) * u.au / u.day
            vz_observer = np.array([svzs[t] for t in alltimes]) * u.au / u.day

        # Only used for the topocentric calculation.
        else:
            if (obscode != 500) and (obscode != '500'):
                earth += wgs84.latlon(obslat, obslon, obselev)

            ee = earth.at(t)

            xx, yy, zz = ee.position.au * u.au # earth ICRS position
            vxx, vyy, vzz = ee.velocity.au_per_d * u.au / u.day # earth ICRS position


            exs = {t:x.value for t, x in zip(unique_times, xx)}
            eys = {t:y.value for t, y in zip(unique_times, yy)}
            ezs = {t:z.value for t, z in zip(unique_times, zz)}
            evxs = {t:vx.value for t, vx in zip(unique_times, vxx)}
            evys = {t:vy.value for t, vy in zip(unique_times, vyy)}
            evzs = {t:vz.value for t, vz in zip(unique_times, vzz)}

            x_observer = np.array([exs[t] for t in alltimes]) * u.au
            y_observer = np.array([eys[t] for t in alltimes]) * u.au
            z_observer = np.array([ezs[t] for t in alltimes]) * u.au

            vx_observer = np.array([evxs[t] for t in alltimes]) * u.au / u.day
            vy_observer = np.array([evys[t] for t in alltimes]) * u.au / u.day
            vz_observer = np.array([evzs[t] for t in alltimes]) * u.au / u.day

        x0, y0, z0 = self.x, self.y, self.z
        vx0, vy0, vz0 = self.vx, self.vy, self.vz

        ltt0 = 0

        N = len(self)
        a = self.a.au.astype(np.double)
        e = self.e.astype(np.double)
        inc = self.inc.rad.astype(np.double)
        arg = self.arg.rad.astype(np.double)
        node = self.node.rad.astype(np.double)


        for idx in range(10):

            dx = x0 - x_observer
            dy = y0 * np.cos(epsilon) - z0 * np.sin(epsilon) - y_observer
            dz = y0 * np.sin(epsilon) + z0 * np.cos(epsilon) - z_observer
            dvx = vx0 - vx_observer
            dvy = vy0 * np.cos(epsilon) - vz0 * np.sin(epsilon) - vy_observer
            dvz = vy0 * np.sin(epsilon) + vz0 * np.cos(epsilon) - vz_observer

            delta = sqrt(dx**2 + dy**2 + dz**2)

            ltt = delta / c
            dltt = (ltt - ltt0)

            if np.all(abs(max(dltt.value)) < 1e-12):
                break

            M = self.M - (ltt * self.n)
            x0, y0, z0, vx0, vy0, vz0 = self.kep_to_xyz_temp(N, a, e, inc, arg, node, M.rad.astype(np.double))

            ltt0 = ltt

        # Be polite
        if in_frame == 'heliocentric':
            self.to_helio()

        return dx, dy, dz, dvx, dvy, dvz


    def set_simulation(self, startdate, units, model):

        sun = planets['sun']
        mercury = planets['mercury barycenter']
        venus = planets['venus barycenter']
        earth = planets['earth']
        moon = planets['moon']
        mars = planets['mars barycenter']
        jupiter = planets['jupiter barycenter']
        saturn = planets['saturn barycenter']
        uranus = planets['uranus barycenter']
        neptune = planets['neptune barycenter']
        pluto = planets['pluto barycenter']

        M_sun = 1
        M_mercury = 1.6601367952719304e-7 # Mercury Barycenter
        M_venus = 2.4478383396645447e-6 # Venus Barycenter
        M_earth = 3.0034896161241036e-06
        M_moon = M_earth / 81.3005690769
        M_mars = 3.2271560375549977e-7 # Mars Barycenter

        M_jupiter = 9.547919384243222e-4
        M_saturn = 2.858859806661029e-4
        M_uranus = 4.3662440433515637e-5
        M_neptune = 5.151389020535497e-5
        M_pluto = 7.361781606089469e-9


        if model == 0:
            active_bodies = [sun]
            names = ['Sun']
            M_sun += M_mercury + M_venus + M_earth + M_mars \
                     + M_jupiter + M_saturn + M_uranus + M_neptune
            masses = [M_sun]


        elif model == 1:
            active_bodies = [sun, jupiter, saturn, uranus, neptune]
            names = ['Sun', 'Jupiter', 'Saturn', 'Uranus', 'Neptune']
            M_sun += M_mercury + M_venus + M_earth + M_mars
            masses = [M_sun, M_jupiter, M_saturn, M_uranus, M_neptune]


        elif model == 2:
            active_bodies = [sun, mercury, venus, earth, moon, mars, jupiter, saturn, uranus, neptune, pluto]
            names = ['Sun', 'Mercury', 'Venus', 'Earth', 'Moon', 'Mars', 'Jupiter', 'Saturn', 'Uranus', 'Neptune', 'Pluto']
            masses = [M_sun, M_mercury, M_venus, M_earth, M_moon, M_mars, M_jupiter, M_saturn, M_uranus, M_neptune, M_pluto]

        else:
            raise ValueError('Model not recognized. Check the documentation.')


        startdate = Time(startdate, scale=units.timescale, format='jd')
        t = ts.tdb(jd=startdate.tdb.jd)

        x, y, z = np.array([body.at(t).ecliptic_xyz().au for body in active_bodies]).T
        vx, vy, vz = np.array([body.at(t).ecliptic_velocity().au_per_d for body in active_bodies]).T


        # create a dataframe of the massive bodies in the solar system
        ss = pd.DataFrame()
        ss['x'] = x
        ss['y'] = y
        ss['z'] = z
        ss['vx'] = vx
        ss['vy'] = vy
        ss['vz'] = vz
        ss['mass'] = masses
        ss['a'] = 1 / (2 / sqrt(ss.x**2 + ss.y**2 + ss.z**2) - (ss.vx**2 + ss.vy**2 + ss.vz**2) / mu_bary.value)
        ss['hill_radius'] = ss.a * pow(ss.mass / (3 * M_sun), 1/3)
        ss['name'] = names


        sim = rebound.Simulation()
        sim.units = ('day', 'AU', 'Msun')

        for p in ss.itertuples():
            sim.add(x=p.x, y=p.y, z=p.z,
                    vx=p.vx, vy=p.vy, vz=p.vz,
                    m=p.mass, hash=p.name, r=p.hill_radius)

        sim.N_active = len(ss)

        #if gr == True:
        #    rebx = reboundx.Extras(sim)
        #    gr = rebx.load_force('gr_full')
        #    gr.params["c"] = constants.C
        #    rebx.add_force(gr)

        sim.testparticle_type = 0
        sim.integrator = 'ias15'

        #sim.move_to_com()

        return sim, names


    def orbits(self):

        M = Angle(np.linspace(0, 2*np.pi, 1000), u.rad)

        xs = []
        ys = []
        zs = []

        for r in self:
            x, y, z, _, _, _ = self.kep_to_xyz_temp(np.repeat(r.a, 1000),
                                                    np.repeat(r.e, 1000),
                                                    np.repeat(r.inc, 1000),
                                                    np.repeat(r.arg, 1000),
                                                    np.repeat(r.node, 1000),
                                                    M)
            xs.append(x)
            ys.append(y)
            zs.append(z)

        return xs, ys, zs

    def kep_to_xyz_temp(self, N, a, e, inc, arg, node, M):

        rock = clibspacerocks.kep_to_xyz_temp(N, a, e, inc, arg, node, M)
        arr = np.ctypeslib.as_array(rock, (6 * N,))

        x, y, z, vx, vy, vz = arr.reshape(N, 6).T
        return x * u.au, y * u.au, z * u.au, vx * u.au/u.d, vy * u.au/u.d, vz * u.au/u.d



    #def kep_to_xyz_temp(self, a, e, inc, arg, node, M):
    #    '''
    #    Just compute the xyz position of an object. Used for iterative
    #    equatorial calculation.
    #    '''

    #    E = np.zeros(len(M))

    #    MM = M[e < 1].rad
    #    ee = e[e < 1]

    #    MM[MM > pi] -= 2 * pi
    #    alpha = (3 * pi**2 + 1.6 * (pi**2 - pi * abs(MM))/(1 + ee))/(pi**2 - 6)
    #    d = 3 * (1 - ee) + alpha * ee
    #    q = 2 * alpha * d * (1 - ee) - MM**2
    #    r = 3 * alpha * d * (d - 1 + ee) * MM + MM**3
    #    w = (abs(r) + sqrt(q**3 + r**2))**(2/3)
    #    E1 = (2 * r * w / (w**2 + w*q + q**2) + MM)/d
    #    f2 = ee * sin(E1)
    #    f3 = ee * cos(E1)
    #    f0 = E1 - f2 - MM
    #    f1 = 1 - f3
    #    d3 = -f0 / (f1 - f0 * f2 / (2 * f1))
    #    d4 = -f0 / (f1 + f2 * d3 / 2 + d3**2 * f3 / 6)
    #    d5 = -f0 / (f1 + d4 * f2 / 2 + d4**2 * f3 / 6 - d4**3 * f2 / 24)
    #    E[e < 1] = (E1 + d5) % (2 * pi)

    #    if np.any(e >= 1):
    #        MMM = M[e >= 1].rad
    #        eee = e[e >= 1]
    #        f = lambda E, MMM, eee: eee * sinh(E) - E - MMM
    #        E0 = MMM
    #        E[e >= 1] = np.array([newton(f, E0[idx], args=(MMM[idx], eee[idx]), maxiter=10000) for idx in range(len(MMM))])

    #    # compute true anomaly ν
    #    #true_anomaly = 2 * np.arctan2((1 + e)**0.5*np.sin(E/2), (1 - e)**0.5*np.cos(E/2))

    #    true_anomaly = np.zeros(len(M))
    #    true_anomaly[e < 1] = 2 * arctan2(sqrt(1 + e[e < 1]) * sin(E[e < 1] / 2), sqrt(1 - e[e < 1]) * cos(E[e < 1] / 2))
    #    true_anomaly[e >= 1] = 2 * arctan2(sqrt(e[e >= 1] + 1) * tanh(E[e >= 1] / 2), sqrt(e[e >= 1] - 1))


    #    r = np.zeros(len(M))
    #    r[e < 1] = a[e < 1] * (1 - e[e < 1] * cos(E[e < 1]))
    #    r[e >= 1] = abs(a[e >= 1]) * abs(1 - e[e >= 1]**2) / (1 + e[e >= 1] * cos(true_anomaly[e >= 1]))

    #    # compute the distance to the central body r
    #    #r = a * (1 - e * np.cos(E))

    #    # obtain the position vector o
    #    o = Vector(r * cos(true_anomaly), r * sin(true_anomaly), np.zeros_like(true_anomaly))

    #    vx = np.zeros(len(M))
    #    vy = np.zeros(len(M))
    #    vz = np.zeros(len(M))
    #    xi_bound = sqrt(mu_bary / u.rad**2 * abs(a[e < 1])) / r[e < 1]
    #    xi_unbound = sqrt(mu_bary / u.rad**2 * abs(a[e >= 1])) / r[e >= 1]

    #    vx[e < 1] = - xi_bound * sin(E[e < 1])
    #    vx[e >= 1] = - xi_unbound * sinh(E[e >= 1])

    #    vy[e < 1] = xi_bound * sqrt(abs(1 - e[e < 1]**2)) * cos(E[e < 1])
    #    vy[e >= 1] = xi_unbound * sqrt(abs(1 - e[e >= 1]**2)) * np.cosh(E[e >= 1])

    #    vz[e < 1] = xi_bound * zeros_like(E[e < 1])
    #    vz[e >= 1] = xi_unbound * zeros_like(E[e >= 1])

    #    ov = Vector(vx, vy, vz)

    #    #ov = Vector((mu_bary * a)**0.5 / r * (-np.sin(E))/ u.rad, (mu_bary * a)**0.5 / r * ((1 - e**2)**0.5 * np.cos(E))/ u.rad, np.zeros(len(true_anomaly))/ u.rad)

    #    # Rotate o to the inertial frame
    #    position = o.euler_rotation(arg, inc, node) #* u.au
    #    velocity = ov.euler_rotation(arg, inc, node) #* u.au / u.day

    #    return position.x * u.au, position.y * u.au, position.z * u.au, velocity.x * u.au/u.d, velocity.y * u.au/u.d, velocity.z * u.au/u.d



__version__ = '1.0.19'
__author__ = 'Kevin Napier'

# Find suffix
import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

# Import shared library
import os
import warnings
pymodulepath = os.path.dirname(__file__)
from ctypes import cdll
__libpath__ = pymodulepath+"/../libspacerocks"+suffix
clibspacerocks = cdll.LoadLibrary(__libpath__)

from .spacerock import SpaceRock
from .units import Units
#from .propagate import Propagate
#from .observe import Observe
#from .neptune_resonances import *
#from .linalg3d import *
#from .constants import *
#from .jacobians import *



#import numpy as np
#
#class OrbitFit:
#
#    def __init__(self, detections):
#        pass
#        #if mpc_format == True:
#        #    self.file = file
#
#        #elif spacerocks_format = True:
#
#    '''
#    We'll have 4-5 well-measured quantities:
#    alpha, beta, gamma, alpha_dot, and beta_dot.
#    gamma_dot is unknown, but there is an energy constraint.
#
#    We have well-defined uncertainties for ra, dec, ra rate, and dec rate.
#
#    How do we get uncertainties on distance?
#    I think those come from ra rate and dec rate. Maybe we can write down the
#    distance equation, and then propagate the uncertainties.
#
#
#    '''
#
#    @property
#    def beta(self):
#        return arcsin(cos(obliquity) * sin(dec) - sin(obliquity) * cos(dec) * sin(ra))
#
#    @property
#    def lambda(self):
#        return arctan((cos(obliquity) * cos(dec) * sin(ra) + sin(obliquity) * sin(dec)) / (cos(dec) * cos(ra)))
#
#    @property
#    def theta_x(self):
#        #b =
#        #l =
#        return cos(b) * sin(l - l[0])) / (sin(b[0]) * sin(b) - cos(b[0]) * cos(b) * cos(l - l[0]))
#
#    @property
#    def theta_y(self):
#        #b =
#        #l =
#        return cos(b[0]) * cos(b) - sin(b[0]) * np.sin(b) * cos(l - l[0]))/ (sin(b[0]) * sin(b) + cos(b[0]) * cos(b) * np.cos(l - l[0]))



from numpy import pi, sqrt, cbrt, sin, cos, tan, sinh, tanh, exp, log10, zeros_like, arccos, arctan2, array, any, arctanh
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, Distance
from astropy.time import Time
from scipy.optimize import newton
from .constants import *

from .vector import Vector

from skyfield.api import Topos, Loader
# Load in planets for ephemeride calculation.
load = Loader('./Skyfield-Data', expire=False, verbose=False)
ts = load.timescale()
planets = load('de440.bsp')
sun = planets['sun']


class OrbitFuncs:


    def to_bary(self):
        '''
        Method to convert heliocentric coordinates to barycentric coordinates.
        '''
        if self.frame == 'heliocentric':

            t = ts.tdb(jd=self.epoch.tdb.jd)

            x_sun, y_sun, z_sun = sun.at(t).ecliptic_xyz().au * u.au
            vx_sun, vy_sun, vz_sun = sun.at(t).ecliptic_velocity().au_per_d * u.au / u.day

            # calculate the barycentric xyz postion
            x = self.x + x_sun
            y = self.y + y_sun
            z = self.z + z_sun
            vx = self.vx + vx_sun
            vy = self.vy + vy_sun
            vz = self.vz + vz_sun


            self.frame = 'barycentric'
            self.mu = mu_bary

            # clear the keplerian variables because they need to be recomputed
            self.clear_kep()

            self.position = Vector(x, y, z)
            self.velocity = Vector(vx, vy, vz)

        return self


    def to_helio(self):
        '''
        Method to convert barycentric coordinates to heliocentric coordinates.
        '''
        if self.frame == 'barycentric':

            t = ts.tdb(jd=self.epoch.tdb.jd)

            x_sun, y_sun, z_sun = sun.at(t).ecliptic_xyz().au * u.au
            vx_sun, vy_sun, vz_sun = sun.at(t).ecliptic_velocity().au_per_d * u.au / u.day

            # calculate the heliocentric xyz postion
            x = self.x - x_sun
            y = self.y - y_sun
            z = self.z - z_sun
            vx = self.vx - vx_sun
            vy = self.vy - vy_sun
            vz = self.vz - vz_sun

            self.frame = 'heliocentric'
            self.mu = mu_helio

            # clear the keplerian variables because they need to be recomputed
            self.clear_kep()

            self.position = Vector(x, y, z)
            self.velocity = Vector(vx, vy, vz)

        return self

    def clear_kep(self):

        to_delete = ['_a', '_e', '_inc', '_arg', '_node', '_varpi', '_M', '_E',
                     '_true_anomaly', '_true_longitude', '_mean_longitude',
                     '_q', '_t_peri', '_b', '_p', '_n', '_Q', '_hill_radius', '_r',
                     '_ovec', '_vovec', '_nvec', '_hvec', '_evec', '_position', '_velocity', '_rrdot',
                     '_x', '_y', '_z', '_vx', '_vy', '_vz']

        for attr in to_delete:
            self.__dict__.pop(attr, None)

    ''' Vector Quantities '''

    @property
    def ovec(self):
        if not hasattr(self, '_ovec'):
            self.ovec =  Vector(self.r * cos(self.true_anomaly), self.r * sin(self.true_anomaly), zeros_like(self.true_anomaly))
        return self._ovec

    @ovec.setter
    def ovec(self, value):
        self._ovec = value

    @ovec.deleter
    def ovec(self):
        del self._ovec

    @property
    def vovec(self):
        if not hasattr(self, '_vovec'):
            vx = np.zeros(len(self))
            vy = np.zeros(len(self))
            vz = np.zeros(len(self))
            xi_bound = sqrt(self.mu / u.rad**2 * abs(self.a[self.e < 1])) / self.r[self.e < 1]
            xi_unbound = sqrt(self.mu / u.rad**2 * abs(self.a[self.e >= 1])) / self.r[self.e >= 1]

            vx[self.e < 1] = - xi_bound * sin(self.E[self.e < 1].rad)
            vx[self.e >= 1] = - xi_unbound * sinh(self.E[self.e >= 1].rad)

            vy[self.e < 1] = xi_bound * sqrt(abs(1 - self.e[self.e < 1]**2)) * cos(self.E[self.e < 1].rad)
            vy[self.e >= 1] = xi_unbound * sqrt(abs(1 - self.e[self.e > 1]**2)) * np.cosh(self.E[self.e > 1].rad)

            vz[self.e < 1] = xi_bound * zeros_like(self.E[self.e < 1].rad)
            vz[self.e >= 1] = xi_unbound * zeros_like(self.E[self.e >= 1].rad)

            self.vovec = Vector(vx * u.au / u.d, vy * u.au / u.d, vz * u.au / u.d)
        return self._vovec

    @vovec.setter
    def vovec(self, value):
        self._vovec = value

    @vovec.deleter
    def vovec(self):
        del self._vovec

    @property
    def nvec(self):
        if not hasattr(self, '_nvec'):
            self.nvec = Vector(-self.hvec.y, self.hvec.x, zeros_like(self.hvec.z))
        return self._nvec

    @nvec.setter
    def nvec(self, value):
        self._nvec = value

    @nvec.deleter
    def nvec(self):
        del self._nvec

    @property
    def evec(self):
        if not hasattr(self, '_evec'):
            self.evec = self.velocity.cross(self.hvec) / (self.mu / u.rad**2)  - self.position / self.position.norm
        return self._evec

    @evec.setter
    def evec(self, value):
        self._evec = value

    @evec.deleter
    def evec(self):
        del self._evec

    @property
    def hvec(self):
        if not hasattr(self, '_hvec'):
            self.hvec = self.position.cross(self.velocity)
        return self._hvec

    @hvec.setter
    def hvec(self, value):
        self._hvec = value

    @hvec.deleter
    def hvec(self):
        del self._hvec

    @property
    def position(self):
        if not hasattr(self, '_position'):
            if hasattr(self, '_x') and hasattr(self, '_y') and hasattr(self, '_z'):
                self.position = Vector(self.x, self.y, self.z)
            else:
                self.position = self.ovec.euler_rotation(self.arg, self.inc, self.node)
        return self._position

    @position.setter
    def position(self, value):
        self._position = value

    @position.deleter
    def position(self):
        del self._position

    @property
    def velocity(self):
        if not hasattr(self, '_velocity'):
            if hasattr(self, '_vx') and hasattr(self, '_vy') and hasattr(self, '_vz'):
                self.velocity = Vector(self.vx, self.vy, self.vz)
            else:
                self.velocity = self.vovec.euler_rotation(self.arg, self.inc, self.node)
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        self._velocity = value

    @velocity.deleter
    def velocity(self):
        del self._velocity

    ''' Cartesian Elements '''

    @property
    def x(self):
        if not hasattr(self, '_x'):
            self.x = self.position.x
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @x.deleter
    def x(self):
        del self._x

    @property
    def y(self):
        if not hasattr(self, '_y'):
            self.y = self.position.y
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    @y.deleter
    def y(self):
        del self._y

    @property
    def z(self):
        if not hasattr(self, '_z'):
            self.z = self.position.z
        return self._z

    @z.setter
    def z(self, value):
        self._z = value

    @z.deleter
    def z(self):
        del self._z

    @property
    def vx(self):
        if not hasattr(self, '_vx'):
            self.vx = self.velocity.x
        return self._vx

    @vx.setter
    def vx(self, value):
        self._vx = value

    @vx.deleter
    def vx(self):
        del self._vx

    @property
    def vy(self):
        if not hasattr(self, '_vy'):
            self.vy = self.velocity.y
        return self._vy

    @vy.setter
    def vy(self, value):
        self._vy = value

    @vy.deleter
    def vy(self):
        del self._vy

    @property
    def vz(self):
        if not hasattr(self, '_vz'):
            self.vz = self.velocity.z
        return self._vz

    @vz.setter
    def vz(self, value):
        self._vz = value

    @vz.deleter
    def vz(self):
        del self._vz



    ''' Required Keplerian Elements '''

    @property
    def a(self):
        if not hasattr(self, '_a'):
            if hasattr(self, '_position') and hasattr(self, '_velocity'):
                self.a = Distance(1 / (2 / self.position.norm - self.velocity.dot(self.velocity) / self.mu * u.rad**2), u.au, allow_negative=True)
            elif hasattr(self, '_e') and hasattr(self, '_q'):
                self.a = self.q / (1 - self.e)
            elif hasattr(self, '_e') and hasattr(self, '_b'):
                self.a = self.b / np.sqrt(abs(1 - self.e**2)) * (-1 * (self.e > 1))
            elif hasattr(self, '_Q') and hasattr(self, '_q'):
                self.a = self.q / (1 - self.e)
            elif hasattr(self, '_v_inf'):
                self.a = Distance(-self.mu / self.v_inf**2 / u.rad**2, u.au, allow_negative=True)
        return self._a

    @a.setter
    def a(self, value):
        self._a = value

    @a.deleter
    def a(self):
        del self._a


    @property
    def e(self):
        if not hasattr(self, '_e'):

            if hasattr(self, '_position') and hasattr(self, '_velocity'):
                self.e = self.evec.norm.value

            elif hasattr(self, '_a') and hasattr(self, '_q'):
                self.e = 1 - self.q.au / self.a.au

            elif hasattr(self, '_Q') and hasattr(self, '_q'):
                self.e = (1 - self.q.au/self.Q.au) / (1 + self.q.au/self.Q.au)

            elif hasattr(self, '_b') and hasattr(self, '_v_inf'):
                self.e = np.sqrt(1 + self.b.au**2 / self.a.au**2)

        return self._e

    @e.setter
    def e(self, value):
        self._e = value

    @e.deleter
    def e(self):
        del self._e


    @property
    def inc(self):
        if not hasattr(self, '_inc'):
            self.inc = Angle(arccos(self.hvec.z / self.hvec.norm), u.rad)
        return self._inc

    @inc.setter
    def inc(self, value):
        self._inc = value

    @inc.deleter
    def inc(self):
        del self._inc


    ''' Keplerian choice-of-three '''


    @property
    def node(self):
        if not hasattr(self, '_node'):
            if hasattr(self, '_varpi') and hasattr(self, '_arg'):
                self.node = Angle((self.varpi.rad - self.arg.rad) % (2 * pi), u.rad)
            elif hasattr(self, '_position') and hasattr(self, '_velocity'):
                node = zeros_like(self.inc.rad * u.rad)
                node[self.inc == 0] = 0
                node[self.inc != 0] = arccos(self.nvec[self.inc != 0].x / self.nvec[self.inc != 0].norm)
                node[self.nvec.y < 0] = 2 * pi * u.rad - node[self.nvec.y < 0]
                self.node = Angle(node, u.rad)
        return self._node

    @node.setter
    def node(self, value):
        self._node = value

    @node.deleter
    def node(self):
        del self._node


    @property
    def arg(self):
        if not hasattr(self, '_arg'):

            if hasattr(self, '_varpi') and hasattr(self, 'node'):
                self.arg = Angle((self.varpi.rad - self.node.rad) % (2 * pi), u.rad)

            elif hasattr(self, '_position') and hasattr(self, '_velocity'):
                n = self.nvec.norm
                arg = zeros_like(self.e * u.rad)
                arg[(self.e == 0) | (n == 0)] = 0
                arg[(self.e != 0) & (n != 0)] = arccos(self.nvec[(self.e != 0) & (n != 0)].dot(self.evec[(self.e != 0) & (n != 0)]) / (n[(self.e != 0) & (n != 0)] * self.e[(self.e != 0) & (n != 0)]))
                arg[self.evec.z < 0] = 2 * pi * u.rad - arg[self.evec.z < 0]
                self.arg = Angle(arg, u.rad)

        return self._arg

    @arg.setter
    def arg(self, value):
        self._arg = value

    @arg.deleter
    def arg(self):
        del self._arg


    @property
    def varpi(self):
        if not hasattr(self, '_varpi'):
            self.varpi = Angle((self.node.rad + self.arg.rad) % (2 * pi), u.rad)
        return self._varpi

    @varpi.setter
    def varpi(self, value):
        self._varpi = value

    @varpi.deleter
    def varpi(self):
        del self._varpi


    '''

    Keplerian choice-of-six

    There are several options for describing the location of a rock on its orbit:
    - mean anomaly
    - true anomaly
    - eccentric anomaly
    - mean longitude
    - true longitude
    - time of pericenter

    As long as you know the longitude of pericenter, you can get any of these
    quantities from any of the others.

    '''

    @property
    def M(self):
        if not hasattr(self, '_M'):

            if hasattr(self, '_mean_longitude'):
                M = (self.mean_longitude.rad - self.varpi.rad) % (2 * pi)
                self.M = Angle(M, u.rad)

            elif hasattr(self, '_true_anomaly') or hasattr(self, '_true_longitude') or hasattr(self, '_E') or (hasattr(self, '_position') and hasattr(self, '_velocity')):
                M = np.zeros(len(self))
                M[self.e < 1] = (self.E.rad[self.e < 1] - self.e[self.e < 1] * sin(self.E.rad[self.e < 1])) % (2 * pi)
                M[self.e >= 1] = (self.e[self.e >= 1] * sinh(self.E.rad[self.e >= 1]) - self.E.rad[self.e >= 1])
                self.M = Angle(M, u.rad)

            elif hasattr(self, '_t_peri'):
                M = self.n * (self.epoch - self.t_peri)
                M = M % (2 * pi * u.rad)
                self.M = Angle(M, u.rad)

        return self._M

    @M.setter
    def M(self, value):
        self._M = value

    @M.deleter
    def M(self):
        del self._M



    @property
    def E(self):
        if not hasattr(self, '_E'):

            if hasattr(self, '_true_anomaly') or hasattr(self, '_true_longitude'): #or (hasattr(self, '_position') and hasattr(self, '_velocity')):
                E = np.zeros(len(self))
                E[self.e < 1] = 2 * arctan2(sqrt(1 - self.e[self.e < 1]) * sin(self.true_anomaly[self.e < 1].rad/2), sqrt(1 + self.e[self.e < 1]) * cos(self.true_anomaly[self.e < 1].rad/2))
                #E[self.e > 1] = 2 * arctanh(sqrt((self.e[self.e > 1] - 1) / (1 + self.e[self.e > 1])) * tan(self.true_anomaly[self.e > 1].rad/2))
                E[self.e >= 1] = np.arccosh((np.cos(self.true_anomaly[self.e > 1]) + self.e[self.e > 1]) / (1 + self.e[self.e > 1] * np.cos(self.true_anomaly[self.e > 1])))
                self.E = Angle(E, u.rad)

            if (hasattr(self, '_position') and hasattr(self, '_velocity')):
                E = np.zeros(len(self))
                E[self.e < 1] = np.arccos((1 - self.r[self.e < 1]/self.a[self.e < 1]) / self.e[self.e < 1])
                #E[self.e > 1] = 2 * arctanh(sqrt((self.e[self.e > 1] - 1) / (1 + self.e[self.e > 1])) * tan(self.true_anomaly[self.e > 1].rad/2))
                E[self.e >= 1] = np.arccosh((1 - self.r[self.e >= 1]/self.a[self.e >= 1])/self.e[self.e >= 1])
                #E[(self.e >= 1) * (self.rrdot < 0)] *= -1
                E[self.rrdot < 0] *= -1
                self.E = Angle(E, u.rad)


            elif hasattr(self, '_M') or hasattr(self, '_mean_longitude') or hasattr(self, '_t_peri'):
                E = np.zeros(len(self))


                M = array(self.M.rad)[self.e < 1]
                e = self.e[self.e < 1]
                M[M > pi] -= 2 * pi
                alpha = (3 * pi**2 + 1.6 * (pi**2 - pi * abs(M))/(1 + e))/(pi**2 - 6)
                d = 3 * (1 - e) + alpha * e
                q = 2 * alpha * d * (1 - e) - M**2
                r = 3 * alpha * d * (d - 1 + e) * M + M**3
                w = (abs(r) + sqrt(q**3 + r**2))**(2/3)
                E1 = (2 * r * w / (w**2 + w*q + q**2) + M)/d
                f2 = e * sin(E1)
                f3 = e * cos(E1)
                f0 = E1 - f2 - M
                f1 = 1 - f3
                d3 = -f0 / (f1 - f0 * f2 / (2 * f1))
                d4 = -f0 / (f1 + f2 * d3 / 2 + d3**2 * f3 / 6)
                d5 = -f0 / (f1 + d4 * f2 / 2 + d4**2 * f3 / 6 - d4**3 * f2 / 24)
                E[self.e < 1] = (E1 + d5) % (2 * pi)

                if np.any(self.e >= 1):
                    M = array(self.M.rad)[self.e >= 1]
                    e = self.e[self.e >= 1]
                    f = lambda E, M, e: e * sinh(E) - E - M
                    E0 = M
                    E[self.e >= 1] = np.array([newton(f, E0[idx], args=(M[idx], e[idx]), maxiter=10000) for idx in range(len(M))])

                self.E = Angle(E, u.rad)

        return self._E

    @E.setter
    def E(self, value):
        self._E = value

    @E.deleter
    def E(self):
        del self._E


    @property
    def true_anomaly(self):
        if not hasattr(self, '_true_anomaly'):

            if hasattr(self, '_true_longitude'):
                self.true_anomaly = Angle(self.true_longitude.rad - self.varpi.rad, u.rad)

            elif hasattr(self, '_E') or hasattr(self, '_M') or hasattr(self, '_mean_longitude') or hasattr(self, '_t_peri'):
                true_anomaly = np.zeros(len(self))
                true_anomaly[self.e < 1] = 2 * arctan2(sqrt(1 + self.e[self.e < 1]) * sin(self.E.rad[self.e < 1] / 2), sqrt(1 - self.e[self.e < 1]) * cos(self.E.rad[self.e < 1] / 2))
                true_anomaly[self.e >= 1] = 2 * arctan2(sqrt(self.e[self.e >= 1] + 1) * tanh(self.E.rad[self.e >= 1] / 2), sqrt(self.e[self.e >= 1] - 1))
                self.true_anomaly = Angle(true_anomaly, u.rad)
                #self.true_anomaly[(self.e > 1) & (self.rrdot < 0)] *= -1

            elif hasattr(self, '_position') and hasattr(self, '_velocity'):
                true_anomaly = zeros_like(self.e * u.rad)
                true_anomaly[self.e == 0] = arccos(self.position[self.e == 0].x / self.position[self.e == 0].norm)
                true_anomaly[(self.e > 0) * (self.e < 1)] = arccos(self.evec[(self.e > 0) * (self.e < 1)].dot(self.position[(self.e > 0) * (self.e < 1)]) / (self.evec[(self.e > 0) * (self.e < 1)].norm * self.position[(self.e > 0) * (self.e < 1)].norm))
                true_anomaly[self.e >= 1] = 2 * arctan2(sqrt(self.e[self.e >= 1] + 1) * tanh(self.E[self.e >= 1].rad / 2), sqrt(self.e[self.e >= 1] - 1)) * u.rad
                #l = self.a[self.e >= 1] * (self.e[self.e >= 1]**2 - 1)
                #true_anomaly[self.e >= 1] = arccos(((l / self.r[self.e >= 1]) - 1) / self.e[self.e >= 1])
                true_anomaly[self.position.dot(self.velocity).value < 0] = 2 * pi * u.rad - true_anomaly[self.position.dot(self.velocity).value < 0]
                self.true_anomaly = Angle(true_anomaly.value, u.rad) # compute true anomaly from cartesian vectors. Other properties can be computed lazily.

        return self._true_anomaly

    @true_anomaly.setter
    def true_anomaly(self, value):
        self._true_anomaly = value

    @true_anomaly.deleter
    def true_anomaly(self):
        del self._true_anomaly


    @property
    def true_longitude(self):
        if not hasattr(self, '_true_longitude') or (hasattr(self, '_position') and hasattr(self, '_velocity')):
            self.true_longitude = Angle((self.true_anomaly.rad + self.varpi.rad) % (2 * pi), u.rad)
        return self._true_longitude

    @true_longitude.setter
    def true_longitude(self, value):
        self._true_longitude = value

    @true_longitude.deleter
    def true_longitude(self):
        del self._true_longitude

    @property
    def mean_longitude(self):
        if not hasattr(self, '_mean_longitude'):
            self.mean_longitude = Angle((self.M.rad + self.varpi.rad) % (2 * pi), u.rad)
        return self._mean_longitude

    @mean_longitude.setter
    def mean_longitude(self, value):
        self._mean_longitude = value

    @mean_longitude.deleter
    def mean_longitude(self):
        del self._mean_longitude


    @property
    def t_peri(self):
        if not hasattr(self, '_t_peri'):
            self.t_peri = Time(self.epoch.jd * u.day - self.M / self.n, format='jd', scale='utc')
        return self._t_peri

    @t_peri.setter
    def t_peri(self, value):
        self._t_peri = value

    @t_peri.deleter
    def t_peri(self):
        del self._t_peri



    ''' Physical Properties '''

    #@property
    #def mass(self):
    #    if not hasattr(self, '_mass'):
    #        self.mass = 0
    #    return self._mass

    #@mass.setter
    #def mass(self, value):
    #    self._mass = value


    #@property
    #def radius(self):
    #    if not hasattr(self, '_radius'):
    #        self.radius = 1e-16
    #    return self._radius

    #@radius.setter
    #def radius(self, value):
    #    self._radius = value


    #@property
    #def density(self):
    #    if not hasattr(self, '_density'):
    #        self.density = 1e-16
    #    return self._density

    #@density.setter
    #def density(self, value):
    #    self._density = value


    ''' Derived Quantities '''

    @property
    def b(self):
        if not hasattr(self, '_b'):
            self.b = abs(self.a) * sqrt(abs(1 - self.e**2))
        return self._b

    @b.setter
    def b(self, value):
        self._b = value

    @b.deleter
    def b(self):
        del self._b


    @property
    def p(self):
        if not hasattr(self, '_p'):
            self.p = abs(self.a) * abs(1 - self.e**2)
        return self._p

    @p.setter
    def p(self, value):
        self._p = value

    @p.deleter
    def p(self):
        del self._p


    @property
    def q(self):
        if not hasattr(self, '_q'):
            self.q = self.a * (1 - self.e)
        return self._q

    @q.setter
    def q(self, value):
        self._q = value

    @q.deleter
    def q(self):
        del self._q


    @property
    def Q(self):
        if not hasattr(self, '_Q'):
            self.Q = self.a * (1 + self.e)
        return self._Q

    @Q.setter
    def Q(self, value):
        self._Q = value

    @Q.deleter
    def Q(self):
        del self._Q


    @property
    def v_inf(self):
        if not hasattr(self, '_v_inf'):
            self.v_inf = sqrt(- self.mu / self.a) / u.rad
        return self._v_inf

    @v_inf.setter
    def v_inf(self, value):
        self._v_inf = value

    @v_inf.deleter
    def v_inf(self):
        del self._v_inf


    @property
    def n(self):
        if not hasattr(self, '_n'):
            self.n = sqrt(self.mu / abs(self.a)**3)
        return self._n

    @n.setter
    def n(self, value):
        self._n = value

    @n.deleter
    def n(self):
        del self._n

    @property
    def rrdot(self):
        '''Calculate the dot product of the position and velocity vectors'''
        return self.position.dot(self.velocity)

    @property
    def r(self):
        '''Compute the distance of the rock from the center of the coordinate system'''
        if not hasattr(self, '_r'):
            if hasattr(self, '_position'):
                self.r = Distance(self.position.norm)
            else:
                rs = np.zeros(len(self))
                rs[self.e < 1] = self.a[self.e < 1] * (1 - self.e[self.e < 1] * cos(self.E[self.e < 1].rad))
                rs[self.e >= 1] = self.p[self.e >=1] / (1 + self.e[self.e >= 1] * cos(self.true_anomaly[self.e >= 1]))
                self.r = Distance(rs, u.au)
        return self._r

    @r.setter
    def r(self, value):
        self._r = value

    @r.deleter
    def r(self):
        del self._r


    #@property
    #def mag(self):
    #    if not hasattr(self, '_mag'):
    #        self.mag = 0
    #    return self._mag

    #@mag.setter
    #def mag(self, value):
    #    self._mag = value

    #@mag.deleter
    #def mag(self):
    #    del self._mag


    #@property
    #def H(self):
    #    if not hasattr(self, '_H'):
    #        if hasattr(self, '_mag'):
    #            obs = self.observe()
    #            self.H = 0
    #    return self._H

    #@H.setter
    #def H(self, value):
    #    self._H = value

    #@H.deleter
    #def H(self):
    #    del self._mag



    #@property
    #def hill_radius(self):
    #    if not hasattr(self, '_hill_radius'):
    #        self.hill_radius = self.a * (1 - self.e) * cbrt(self.m / 3)
    #    return self._hill_radius

    #@hill_radius.setter
    #def hill_radius(self, value):
    #    self._hill_radius = value

    #@hill_radius.deleter
    #def hill_radius(self):
    #    del self._hill_radius

    @property
    def TisserandJ(self):
        aJ = 5.2 * u.au
        return aJ / self.a * 2 * cos(self.inc) * sqrt(self.p / aJ)



import random
import copy
import os

from astropy import units as u
from astropy.coordinates import Angle, Distance
from astropy.time import Time

from numpy import sin, cos, arctan2, sqrt, array, pi, zeros, arcsin, tan
import numpy as np
import pandas as pd

from .constants import c, epsilon

from skyfield.api import Topos, Loader
# Load in planets for ephemeride calculation.
load = Loader('./Skyfield-Data', expire=False, verbose=False)
ts = load.timescale()
planets = load('de440.bsp')
earth = planets['earth']
sun = planets['sun']

from .convenience import Convenience

class Ephemerides(Convenience):

    '''
    Take in the ecliptic state vector (relative to the observer) and
    return any quantity
    '''

    def __init__(self, **kwargs):
        self.x = kwargs.get('x')
        self.y = kwargs.get('y')
        self.z = kwargs.get('z')
        self.vx = kwargs.get('vx')
        self.vy = kwargs.get('vy')
        self.vz = kwargs.get('vz')
        self.epoch = kwargs.get('epoch')
        self.name = kwargs.get('name')

        if kwargs.get('H0') is not None:
            self.H0 = kwargs.get('H0')
            self.G = kwargs.get('G')

        if kwargs.get('r_helio') is not None:
            self.r_helio = kwargs.get('r_helio')

        if kwargs.get('delta_H') is not None:
            self.delta_H = kwargs.get('delta_H')
            self.rotation_period = kwargs.get('rotation_period')
            self.phi0 = kwargs.get('phi0')
            self.t0 = kwargs.get('t0')

        self.delta = Distance(np.sqrt(self.x**2 + self.y**2 + self.z**2), u.au)



    @property
    def ra(self):
        if not hasattr(self, '_ra'):
            self.ra = Angle(np.arctan2(self.y, self.x), u.rad).wrap_at(2 * np.pi * u.rad)
        return self._ra

    @ra.setter
    def ra(self, value):
        self._ra = value

    @ra.deleter
    def ra(self):
        del self._ra

    @property
    def dec(self):
        if not hasattr(self, '_dec'):
            self.dec = Angle(np.arcsin(self.z / sqrt(self.x**2 + self.y**2 + self.z**2)), u.rad)
        return self._dec

    @dec.setter
    def dec(self, value):
        self._dec = value

    @dec.deleter
    def dec(self):
        del self._dec

    @property
    def ra_rate(self):
        if not hasattr(self, '_ra_rate'):
            self.ra_rate = -cos(self.dec) * (self.y * self.vx - self.x * self.vy) / (self.x**2 + self.y**2) * u.rad
        return self._ra_rate

    @ra_rate.setter
    def ra_rate(self, value):
        self._ra_rate = value

    @ra_rate.deleter
    def ra_rate(self):
        del self._ra_rate

    @property
    def dec_rate(self):
        if not hasattr(self, '_dec_rate'):
            self.dec_rate = (-self.z * (self.x * self.vx + self.y * self.vy) + ((self.x**2 + self.y**2) * self.vz)) \
                            / (sqrt(self.x**2 + self.y**2) * (self.x**2 + self.y**2 + self.z**2)) * u.rad
        return self._dec_rate

    @dec_rate.setter
    def dec_rate(self, value):
        self._dec_rate = value

    @dec_rate.deleter
    def dec_rate(self):
        del self._dec_rate

    @property
    def mag(self):
        if not hasattr(self, '_mag'):
            self.mag = self.estimate_mag()
        return self._mag

    @mag.setter
    def mag(self, value):
        self._mag = value

    @mag.deleter
    def mag(self):
        del self._mag


    @property
    def H(self):
        if not hasattr(self, '_H'):
            if hasattr(self, 'delta_H'):
                ltt = self.delta / c
                dH = self.delta_H * np.sin((self.epoch.jd - self.t0.jd - ltt.value) * 2 * np.pi / self.rotation_period + self.phi0.rad)

                self.H = self.H0 + dH

            else:
                self.H = self.H0
        return self._H

    @H.setter
    def H(self, value):
        self._H = value

    @H.deleter
    def H(self):
        del self._H


    @property
    def b(self):
        return Angle(arcsin(cos(epsilon) * sin(self.dec) - sin(epsilon) * cos(self.dec) * sin(self.ra)), u.rad)

    @property
    def l(self):
        return Angle(arctan2((cos(epsilon) * cos(self.dec) * sin(self.ra) + sin(epsilon) * sin(self.dec)), (cos(self.dec) * cos(self.ra))), u.rad)

    @property
    def b_rate(self):
        num = self.dec_rate * (cos(epsilon) * cos(self.dec) + sin(epsilon) * sin(self.dec) * sin(self.ra)) - self.ra_rate * cos(self.dec) * cos(self.ra) * sin(epsilon)
        denom = sqrt(1 - (cos(epsilon) * sin(self.dec) - cos(self.dec) * sin(self.ra) * sin(epsilon))**2)
        return num / denom

    @property
    def l_rate(self):
        num = self.ra_rate * (1/cos(self.ra)) * (cos(epsilon) * (1/cos(self.ra)) + sin(epsilon) * tan(self.ra))
        denom = 1 + ((1/cos(self.ra)) * sin(epsilon) + cos(epsilon) * tan(self.ra))**2
        return num / denom




    def estimate_mag(self):
        '''
        Estimate the apparent magnitude of a TNO
        https://iopscience.iop.org/article/10.3847/1538-3881/ab18a9/pdf for light curves
        '''
        t = ts.tdb(jd=self.epoch.tdb.jd)
        e = earth.at(t)
        s = sun.at(t)
        sx, sy, sz = s.ecliptic_xyz().au
        ex, ey, ez = e.ecliptic_xyz().au
        earth_dist = ((ex-sx)**2 + (ey-sy)**2 + (ez-sz)**2)**0.5

        q = (self.r_helio.au**2 + self.delta.au**2 - earth_dist)/(2 * self.r_helio.au * self.delta.au)

        ## pyephem
        beta = np.arccos(q)
        beta[np.where(q <= -1)[0]] = np.pi * u.rad
        beta[np.where(q >= 1)[0]] = 0 * u.rad

        Psi_1 = np.exp(-3.332 * np.tan(beta/2)**0.631)
        Psi_2 = np.exp(-1.862 * np.tan(beta/2)**1.218)
        mag = self.H + 5 * np.log10(self.r_helio.au * self.delta.au)# / earth_dist**2)


        not_zero = np.where((Psi_1 != 0) | (Psi_2 != 0))[0]
        mag[not_zero] -= 2.5 * np.log10((1 - self.G[not_zero]) * Psi_1[not_zero] + self.G[not_zero] * Psi_2[not_zero])

        return mag



#    def plot_radec(self, color='black', alpha=0.5, zoom=False, galactic_plane=False, ecliptic_plane=True):
#        '''
#        Plot the right ascension and declination of each object on a Mollweide
#        projection. (See https://en.wikipedia.org/wiki/Mollweide_projection)
#        '''
#        if zoom == True:
#
#            fig = plt.figure(figsize=(12, 8))
#            ax = fig.add_subplot(111, projection=ccrs.PlateCarree())
#
#            xdata = self.ra.degree
#            xdata[xdata > 180] -= 360
#
#            xmin=np.min(xdata)
#            xmax=np.max(xdata)
#            ymin=np.min(self.dec.degree)
#            ymax=np.max(self.dec.degree)
#
#            ax.scatter(-xdata, self.dec.degree, color='black', alpha=0.5)
#
#            xticks = np.linspace(-xmax, -xmin, 8)
#            yticks = np.linspace(ymin, ymax, 8)
#
#            ax.set_xticks(xticks)
#            ax.set_yticks(yticks)
#            xticklabels = [r'${:.2f}\degree$'.format(-value) for value in xticks]
#            yticklabels = [r'${:.2f}\degree$'.format(value) for value in yticks]
#            ax.set_xticklabels(xticklabels)
#            ax.set_yticklabels(yticklabels)
#
#            gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
#                              linewidth=1, color='gray', alpha=0.5, linestyle='-')
#            gl.xlocator = mticker.FixedLocator(xticks)
#            gl.ylocator = mticker.FixedLocator(yticks)
#
#            gl.bottom_labels = False
#            gl.top_labels = False
#            gl.left_labels = False
#            gl.right_labels = False
#
#            xrange = xmax - xmin
#            yrange = ymax - ymin
#
#            try:
#                ax.set_extent([-xmax - xrange * 0.05, -xmin + xrange * 0.05,
#                               ymin - yrange * 0.05, ymax + yrange * 0.05], crs=ccrs.PlateCarree())
#            except:
#                ax.set_extent([-180, 180, -90, 90], crs=ccrs.PlateCarree())
#
#            if ecliptic_plane == True:
#                def radec2project(ra, dec):
#                    ra[ra>180] -= 360
#                    return (ra, dec)
#
#                plane_lon = np.linspace(-np.pi, np.pi, 1000)
#                plane_lat = np.zeros(len(plane_lon))
#                ecl_plane = np.zeros([len(plane_lat), 2])
#
#                for i in range(len(plane_lat)):
#                    ecl_plane[i] = ephem.Equatorial(ephem.Ecliptic(plane_lon[i], plane_lat[i])).get()
#
#                x, y = radec2project(np.degrees(ecl_plane.T[0]), np.degrees(ecl_plane.T[1]))
#                ax.plot(x[1:999], -y[1:999], 'r-', zorder=2)
#
#            if galactic_plane == True:
#
#                galactic = SkyCoord(l=np.linspace(0, 360, 1000)*u.degree, b=np.zeros(1000)*u.degree, frame='galactic')
#                galxdata = galactic.icrs.ra.degree
#                galxdata[galxdata > 180] -= 360
#                galydata = galactic.icrs.dec.degree
#                order = galxdata.argsort()
#                ax.plot(-galxdata[order][1:999], -galydata[order][1:999], 'g-', zorder=3)
#
#        else:
#
#            fig = plt.figure(figsize=(12, 8))
#            ax = fig.add_subplot(111, projection='mollweide')
#            ax.grid(True)
#
#            xdata = self.ra
#            xdata[xdata.value > np.pi] -= 2*np.pi * u.rad
#            ax.set_xticklabels([r'$150\degree$', r'$120\degree$', r'$90\degree$' ,
#                                r'$60\degree$' , r'$30\degree$' , r'$0\degree$'  ,
#                                r'$330\degree$', r'$300\degree$', r'$270\degree$',
#                                r'$240\degree$', r'$210\degree$'])
#
#            ax.scatter(-xdata, self.dec, color=color, alpha=alpha)
#
#            if ecliptic_plane == True:
#                def radec2project(ra, dec):
#                    ra[ra>180] -= 360
#                    return (ra, dec)
#
#                plane_lon = np.linspace(-np.pi, np.pi, 1000)
#                plane_lat = np.zeros(len(plane_lon))
#                ecl_plane = np.zeros([len(plane_lat), 2])
#
#                for i in range(len(plane_lat)):
#                    ecl_plane[i] = ephem.Equatorial(ephem.Ecliptic(plane_lon[i], plane_lat[i])).get()
#
#                x, y = radec2project(np.degrees(ecl_plane.T[0]), np.degrees(ecl_plane.T[1]))
#                ax.plot(np.radians(x[1:999]), -np.radians(y[1:999]), 'r-', zorder=2)
#
#            if galactic_plane == True:
#
#                galactic = SkyCoord(l=np.linspace(0, 360, 1000)*u.degree, b=np.zeros(1000)*u.degree, frame='galactic')
#                galxdata = galactic.icrs.ra.rad
#                galxdata[galxdata > np.pi] -= 2*np.pi
#                galydata = galactic.icrs.dec.rad
#                order = galxdata.argsort()
#                ax.plot(-galxdata[order][1:999], galydata[order][1:999], 'g-', zorder=3)
#
#        return fig, ax
#



from astropy import units as u
from astropy.table import Table

# Units manager for spacerocks.
# Usage:
# from astropy import units as u
# units = Units()
# units.angle = u.rad

class Units:

    def __init__(self):
        self.set_default()

    def set_default(self):
        self.distance = u.au
        self.angle = u.deg
        self.timescale = 'utc'
        self.timeformat = None
        self.speed = u.au / u.d
        self.rotation_period = u.d

    def current(self):
        print("{:<20} {:<15}".format('Quantity', 'Unit'))
        print('---------------------------------------')
        for k, v in self.__dict__.items():
            print("{:<20} {:<15}".format(k, str(v)))



from setuptools import setup, Extension

import os
import sys

import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

extra_link_args=[]
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-shared')
    extra_link_args=['-Wl,-install_name,@rpath/libspacerocks'+suffix]

libspacerocksmodule = Extension('libspacerocks',
                                sources=['src/speedy.c'],
                                include_dirs=['src'],
                                language='c',
                                extra_compile_args=['-O3', '-fPIC', '-Wno-unknown-pragmas', '-std=c99'],
                                extra_link_args=extra_link_args
                                )


setup(
   name='spacerocks',
   version='1.0.19',
   description='A Python Package for Solar System Ephemerides and Dynamics.',
   author='Kevin J. Napier',
   author_email='kjnapier@umich.edu',
   url="https://github.com/kjnapier/spacerocks",
   packages=['spacerocks', 'spacerocks.data'],
   package_data={'spacerocks.data': ['*.csv']},
   include_package_data=True,
   # I'm not sure package_data is working.
   install_requires=['healpy',
                     'numpy',
                     'skyfield',
                     'astropy',
                     'pandas',
                     'rebound',
                     'reboundx'],
   ext_modules=[libspacerocksmodule],
   zip_safe=False
)



from numpy import sqrt, sin, cos
import copy

class Vector:

    def __init__(self, x, y, z):

        self.x = x
        self.y = y
        self.z = z

    def __getitem__(self, idx):
        '''
        This method allows you to index a Vector object.
        '''
        p = copy.copy(self)
        for attr in self.__dict__.keys():
            setattr(p, attr, getattr(self, attr)[idx])

        return p

    def __add__(self, other):
        return Vector(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Vector(self.x - other.x, self.y - other.y, self.z - other.z)

    def __rmul__(self, c):
        return Vector(c * self.x, c * self.y, c * self.z)

    def __truediv__(self, c):
        return Vector(self.x / c, self.y / c, self.z / c)

    def dot(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        x = self.y * other.z - self.z * other.y
        y = self.z * other.x - self.x * other.z
        z = self.x * other.y - self.y * other.x
        return Vector(x, y, z)

    def euler_rotation(self, a, b, c):
        sa = sin(a)
        sb = sin(b)
        sc = sin(c)
        ca = cos(a)
        cb = cos(b)
        cc = cos(c)
        # Currently assumes z=0. I'm going to modify this to the genreal case.
        xrot = self.x * (ca * cc - sa * sc * cb) - self.y * (sa * cc + ca * sc * cb)
        yrot = self.x * (ca * sc + sa * cc * cb) + self.y * (ca * cc * cb - sa * sc)
        zrot = self.x * (sa * sb) + self.y * (ca * sb)
        #xrot = self.x * (cos(a) * cos(c) - sin(a) * sin(c) * cos(b)) - self.y * (sin(a) * cos(c) + cos(a) * sin(c) * cos(b))
        #yrot = self.x * (cos(a) * sin(c) + sin(a) * cos(c) * cos(b)) + self.y * (cos(a) * cos(c) * cos(b) - sin(a) * sin(c))
        #zrot = self.x * (sin(a) * sin(b)) + self.y * (cos(a) * sin(b))
        return Vector(xrot, yrot, zrot)


    @property
    def norm(self):
        return sqrt(self.x**2 + self.y**2 + self.z**2)

    @property
    def unit(self):
        return self / self.norm



from astropy.table import Table
import pandas as pd
import numpy as np
from astropy.time import Time
import copy

import dateutil

class Convenience:

    def __len__(self):
        '''
        This method allows you to use the len() function on a SpaceRocks object.
        '''
        return len(self.name)


    def __getitem__(self, idx):
        '''
        This method allows you to index a SpaceRocks object.
        '''
        p = copy.copy(self)
        for attr in self.__dict__.keys():
            if (attr != 'mu') and (attr != 'frame') and (attr != 'G'):
                setattr(p, attr, getattr(self, attr)[idx])

        return p


    def detect_timescale(self, timevalue, timescale):
        if isinstance(timevalue[0], str):
            dates = [dateutil.parser.parse(d, fuzzy_with_tokens=True)[0] for d in timevalue]
            return Time(dates, format='datetime', scale=timescale)
        elif np.all(timevalue > 100000):
            return Time(timevalue, format='jd', scale=timescale)
        else:
            return Time(timevalue, format='mjd', scale=timescale)

    def detect_coords(self, kwargs):

        kep = ['a', 'e', 'q', 'inc', 'node', 'arg', 'M', 'true_anomaly', 'E', 'varpi', 't_peri', 'mean_longitude', 'true_longitude', 'v_inf', 'b', 'Q']
        xyz = ['x', 'y', 'z', 'vx', 'vy', 'vz']

        input_coords = list(kwargs.keys())

        if np.in1d(input_coords, kep).sum() == 6:
            coords = 'kep'

        elif np.in1d(input_coords, xyz).sum() == 6:
            coords = 'xyz'

        else:
            raise ValueError('Invalid input coordinates. Please see the documentation for accepted input.')

        return coords



    def astropy_table(self):
        '''
        Write the rocks to an astropy table. This can handle units.
        '''
        return Table(self.__dict__)


    def pandas_df(self):
        '''
        Write the rocks to a pandas dataframe. Pandas can't handle astropy
        units, so if you want to keep units intact you'll have to use
        an Astropy Table.
        '''
        return self.astropy_table().to_pandas()
